var getOrCreateClient = require('../client'),
  assert = require('assert'),
  types = {
    ns: require('mongodb-ns'),
    uri: require('mongodb-uri')
  },
  debug = require('debug')('mongoscope-client:backbone');

var _result = function(resource, key){
  return (typeof resource[key] === 'function') ? resource[key]() : resource[key];
};

var _mongodb = function(resource){
  var uri = _result(resource, 'mongodb');
  if(!uri) return undefined;

  if(uri.indexOf('mongodb://') !== 0) uri = 'mongodb://' + uri;

  var info = types.uri.parse(uri), ns = null;

  var mongodb = info.hosts[0].host + ':' + info.hosts[0].port;
  if(info.database && !resource.url){
    ns = types.ns(info.database);
    if(ns.collection) resource.url = '/collections/' + ns.toString() + '/find';
  }
  return mongodb;
};

function getClient(ctx){
  assert(ctx, 'Missing backbone resource');
  return getOrCreateClient({mongodb: _mongodb(ctx)});
}

function sync(method, model, options){
  var url = options.url || _result(this, 'url'),
    client = options.client || getClient(this),
    ender = function(err, res){
      if(err){
        return options.error(err);
      }
      options.success(res);
    };

  // @todo: CRUD is now possible
  if(method !== 'read'){
    throw new Error('mongoscope is readonly, so create, update, ' +
      'patch, and delete sync methods are not available.');
  }

  if(!url){
    throw new Error('A "url" property or function must be specified');
  }

  var params = {}, docs = [];
  Object.keys(options).map(function(k){
    if(['error', 'success', 'client', 'parse'].indexOf(k) === -1){
      params[k] = options[k];
    }
  });

  if(method === 'read'){
    if(!options.all) return client.get(url, params, ender);

    return client.get(url, params)
      .on('error', ender)
      .on('data', function(doc){docs.push(doc);})
      .on('end', function(){
        ender(null, docs);
      });
  }
}

module.exports.Model = {
  sync: sync
};

module.exports.Collection = {
  sync: sync
};

module.exports.ReadableStream = {
  subscription: null,
  subscribe: function(){
    if(this.subscription !== null){
      debug('already have a subscription', this.subscription);
      return this;
    }

    var self = this,
      client = getClient(this),
      url = _result(this, 'url');

    this.subscription = client.get(url)
      .on('error', function(err){
        self.trigger('error', err, self);
      })
      .on('data', function(data){
        if (!self.set(data)) return false;
        self.trigger('sync', self, data);
      });

    // If the client context changes, move our subscription.
    this.subscription.client.on('change', function(){
      self.unsubscribe().subscribe();
    });
    return self;
  },
  unsubscribe: function(){
    if(this.subscription === null){
      debug('already unsubscribed');
      return this;
    }
    this.subscription.close();
    this.subscription = null;
    return this;
  }
};
